# Copyright (C) 2014, 2015 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import functools
import os
import queue
import socket
import signal
import threading
import ipaddress

from virtinst import log

from ..baseclass import vmmGObject


class ConnectionInfo:
    """
    Holds all the bits needed to make a connection to a graphical console
    """

    def __init__(self, conn, gdev):
        self.gtype = gdev.type
        self.gidx = gdev.get_xml_idx()
        self.gport = str(gdev.port) if gdev.port else None
        self.gsocket = (gdev.listens and gdev.listens[0].socket) or gdev.socket
        self.gaddr = gdev.listen or "127.0.0.1"
        self.gtlsport = gdev.tlsPort or None
        self.glistentype = gdev.get_first_listen_type()

        self.transport = conn.get_uri_transport()
        self.connuser = conn.get_uri_username()

        self._connhost = conn.get_uri_hostname() or "localhost"
        self._connport = conn.get_uri_port()
        if self._connhost == "localhost":
            self._connhost = "127.0.0.1"

    def _is_listen_localhost(self, host=None):
        try:
            return ipaddress.ip_network(str(host or self.gaddr)).is_loopback
        except Exception:
            return False

    def _is_listen_any(self):
        try:
            return ipaddress.ip_network(str(self.gaddr)).is_unspecified
        except Exception:
            return False

    def _is_listen_none(self):
        if self.glistentype == "none":
            return True
        return not (self.gsocket or self.gport or self.gtlsport)

    def need_tunnel(self):
        if not self._is_listen_localhost():
            return False
        return self.transport == "ssh"

    def bad_config(self):
        if self.transport and self._is_listen_none():
            return _(
                "Guest is on a remote host, but is only configured "
                "to allow local file descriptor connections."
            )

        if self.need_tunnel() and (self.gtlsport and not self.gport):
            return _("Guest is configured for TLS only which does not work over SSH.")

        if (
            not self.need_tunnel()
            and self.transport
            and self._is_listen_localhost()
            and not self._is_listen_localhost(self._connhost)
        ):
            return (
                _(
                    "Guest is on a remote host with transport '%s' "
                    "but is only configured to listen locally. "
                    "To connect remotely you will need to change the guest's "
                    "listen address."
                )
                % self.transport
            )

    def get_conn_host(self):
        """
        Return host/port/tlsport needed for direct spice/vnc connection
        """
        host = self.gaddr
        port = self.gport
        tlsport = self.gtlsport
        if self._is_listen_any():
            host = self._connhost

        return host, port, tlsport

    def get_tunnel_host(self):
        """
        Physical host name/port needed for ssh tunnel connection
        """
        return self._connhost, self._connport

    def logstring(self):
        return (
            "proto=%s trans=%s connhost=%s connuser=%s "
            "connport=%s gaddr=%s gport=%s gtlsport=%s gsocket=%s"
            % (
                self.gtype,
                self.transport,
                self._connhost,
                self.connuser,
                self._connport,
                self.gaddr,
                self.gport,
                self.gtlsport,
                self.gsocket,
            )
        )


class _TunnelScheduler:
    """
    If the user is using Spice + SSH URI + no SSH keys, we need to
    serialize connection opening otherwise ssh-askpass gets all angry.
    This handles the locking and scheduling.

    It's only instantiated once for the whole app, because we serialize
    independent of connection, vm, etc.
    """

    def __init__(self):
        self._thread = None
        self._queue = queue.Queue()
        self._lock = threading.Lock()

    def _handle_queue(self):
        while True:
            (
                lock_cb,
                cb,
                args,
            ) = self._queue.get()
            lock_cb()
            vmmGObject.idle_add(cb, *args)

    def schedule(self, lock_cb, cb, *args):
        if not self._thread:
            self._thread = threading.Thread(
                name="Tunnel thread", target=self._handle_queue, args=()
            )
            self._thread.daemon = True
        if not self._thread.is_alive():
            self._thread.start()
        self._queue.put((lock_cb, cb, args))

    def lock(self):
        self._lock.acquire()

    def unlock(self):
        self._lock.release()


_tunnel_scheduler = _TunnelScheduler()


class _Tunnel:
    def __init__(self):
        self._pid = None
        self._closed = False
        self._errfd = None

    def close(self):
        if self._closed:
            return  # pragma: no cover
        self._closed = True

        log.debug(
            "Close tunnel PID=%s ERRFD=%s", self._pid, self._errfd and self._errfd.fileno() or None
        )

        # Since this is a socket object, the file descriptor is closed
        # when it's garbage collected.
        self._errfd = None

        if self._pid:
            os.kill(self._pid, signal.SIGKILL)
            os.waitpid(self._pid, 0)
        self._pid = None

    def get_err_output(self):
        errout = ""
        while True:
            try:
                new = self._errfd.recv(1024)
            except Exception:  # pragma: no cover
                break

            if not new:
                break

            errout += new.decode()

        return errout

    def open(self, argv, sshfd):
        if self._closed:
            return  # pragma: no cover

        errfds = socket.socketpair()
        pid = os.fork()
        if pid == 0:  # pragma: no cover
            errfds[0].close()

            os.dup2(sshfd.fileno(), 0)
            os.dup2(sshfd.fileno(), 1)
            os.dup2(errfds[1].fileno(), 2)
            os.execlp(*argv)
            os._exit(1)  # pylint: disable=protected-access

        sshfd.close()
        errfds[1].close()

        self._errfd = errfds[0]
        self._errfd.setblocking(0)
        log.debug("Opened tunnel PID=%d ERRFD=%d", pid, self._errfd.fileno())

        self._pid = pid


def _make_ssh_command(ginfo):
    if not ginfo.need_tunnel():
        return None

    host, port = ginfo.get_tunnel_host()

    # Build SSH cmd
    argv = ["ssh", "ssh"]
    if port:
        argv += ["-p", str(port)]

    if ginfo.connuser:
        argv += ["-l", ginfo.connuser]

    argv += [host]

    # Build 'nc' command run on the remote host
    #
    # This ugly thing is a shell script to detect availability of
    # the -q option for 'nc': debian and suse based distros need this
    # flag to ensure the remote nc will exit on EOF, so it will go away
    # when we close the VNC tunnel. If it doesn't go away, subsequent
    # VNC connection attempts will hang.
    #
    # Fedora's 'nc' doesn't have this option, and apparently defaults
    # to the desired behavior.
    #
    if ginfo.gsocket:
        nc_params = "-U %s" % ginfo.gsocket
    else:
        nc_params = "%s %s" % (ginfo.gaddr, ginfo.gport)

    nc_cmd = (
        """nc -q 2>&1 | grep "requires an argument" >/dev/null;"""
        """if [ $? -eq 0 ] ; then"""
        """   CMD="nc -q 0 %(nc_params)s";"""
        """else"""
        """   CMD="nc %(nc_params)s";"""
        """fi;"""
        """eval "$CMD";""" % {"nc_params": nc_params}
    )

    argv.append("sh -c")
    argv.append("'%s'" % nc_cmd)

    argv_str = functools.reduce(lambda x, y: x + " " + y, argv[1:])
    log.debug("Pre-generated ssh command for ginfo: %s", argv_str)
    return argv


class SSHTunnels:
    def __init__(self, ginfo):
        self._tunnels = []
        self._sshcommand = _make_ssh_command(ginfo)
        self._locked = False

    def open_new(self):
        t = _Tunnel()
        self._tunnels.append(t)

        # socket FDs are closed when the object is garbage collected. This
        # can close an FD behind spice/vnc's back which causes crashes.
        #
        # Dup a bare FD for the viewer side of things, but keep the high
        # level socket object for the SSH side, since it simplifies things
        # in that area.
        viewerfd, sshfd = socket.socketpair()
        _tunnel_scheduler.schedule(self._lock, t.open, self._sshcommand, sshfd)

        retfd = os.dup(viewerfd.fileno())
        log.debug("Generated tunnel fd=%s for viewer", retfd)
        return retfd

    def close_all(self):
        for tunnel in self._tunnels:
            tunnel.close()
        self._tunnels = []
        self.unlock()

    def get_err_output(self):
        errstrings = []
        for tunnel in self._tunnels:
            e = tunnel.get_err_output().strip()
            if e and e not in errstrings:
                errstrings.append(e)
        return "\n".join(errstrings)

    def _lock(self):
        _tunnel_scheduler.lock()
        self._locked = True

    def unlock(self, *args, **kwargs):
        if self._locked:
            _tunnel_scheduler.unlock(*args, **kwargs)
            self._locked = False



# Copyright (C) 2006-2007, 2013 Red Hat, Inc.
# Copyright (C) 2006 Hugh O. Brock <hbrock@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import ipaddress

from gi.repository import Gtk
from gi.repository import Pango

import libvirt

from virtinst import generatename
from virtinst import log
from virtinst import Network

from .lib import uiutil
from .asyncjob import vmmAsyncJob
from .baseclass import vmmGObjectUI
from .xmleditor import vmmXMLEditor


def _make_ipaddr(addrstr):
    try:
        return ipaddress.ip_network(str(addrstr), strict=False)
    except Exception:  # pragma: no cover
        return None


class vmmCreateNetwork(vmmGObjectUI):
    def __init__(self, conn):
        vmmGObjectUI.__init__(self, "createnet.ui", "vmm-create-net")
        self.conn = conn

        self._xmleditor = vmmXMLEditor(
            self.builder, self.topwin, self.widget("net-details-align"), self.widget("net-details")
        )
        self._xmleditor.connect("xml-requested", self._xmleditor_xml_requested_cb)

        self.builder.connect_signals(
            {
                "on_create_cancel_clicked": self.close,
                "on_vmm_create_delete_event": self.close,
                "on_create_finish_clicked": self.finish,
                "on_net_forward_mode_changed": self._net_forward_mode_changed_cb,
                "on_net_forward_device_changed": self._net_forward_device_changed_cb,
                "on_net_dns_use_toggled": self._net_dns_use_toggled_cb,
                "on_net-ipv4-enable_toggled": self._ipv4_toggled_cb,
                "on_net-ipv4-network_changed": self._change_ipv4_network_cb,
                "on_net-dhcpv4-enable_toggled": self._dhcpv4_toggled_cb,
                "on_net-ipv6-enable_toggled": self._ipv6_toggled_cb,
                "on_net-ipv6-network_changed": self._change_ipv6_network_cb,
                "on_net-dhcpv6-enable_toggled": self._dhcpv6_toggled_cb,
            }
        )
        self.bind_escape_key_close()

        self.set_initial_state()

    ####################
    # Standard methods #
    ####################

    def show(self, parent):
        log.debug("Showing new network wizard")
        self.reset_state()
        self.topwin.set_transient_for(parent)
        self.topwin.present()

    def close(self, ignore1=None, ignore2=None):
        log.debug("Closing new network wizard")
        self.topwin.hide()
        return 1

    def _cleanup(self):
        self.conn = None
        self._xmleditor.cleanup()
        self._xmleditor = None

    ###########
    # UI init #
    ###########

    def set_initial_state(self):
        # [ dev name, label ]
        pf_list = self.widget("net-hostdevs")
        pf_model = Gtk.ListStore(str, str)
        pf_list.set_model(pf_model)
        text = uiutil.init_combo_text_column(pf_list, 1)
        text.set_property("ellipsize", Pango.EllipsizeMode.MIDDLE)

        # [ show_manual, label]
        fw_list = self.widget("net-forward-device")
        fw_model = Gtk.ListStore(bool, str)
        fw_list.set_model(fw_model)
        uiutil.init_combo_text_column(fw_list, 1)
        fw_model.append([False, _("Any physical device")])
        fw_model.append([True, _("Physical device...")])

        # [ mode, label ]
        mode_list = self.widget("net-forward-mode")
        mode_model = Gtk.ListStore(str, str)
        mode_list.set_model(mode_model)
        uiutil.init_combo_text_column(mode_list, 1)

        mode_model.append(["nat", _("NAT")])
        mode_model.append(["route", _("Routed")])
        mode_model.append(["open", _("Open")])
        mode_model.append(["isolated", _("Isolated")])
        mode_model.append(["hostdev", _("SR-IOV pool")])

    def reset_state(self):
        self._xmleditor.reset_state()

        basename = "network"

        def cb(n):
            return generatename.check_libvirt_collision(
                self.conn.get_backend().networkLookupByName, n
            )

        default_name = generatename.generate_name(basename, cb)
        self.widget("net-name").set_text(default_name)

        self.widget("net-dns-use-netname").set_active(True)

        self.widget("net-ipv4-expander").set_visible(True)
        self.widget("net-ipv4-expander").set_expanded(False)
        self.widget("net-ipv6-expander").set_visible(True)
        self.widget("net-ipv6-expander").set_expanded(False)
        self.widget("net-dns-expander").set_visible(True)
        self.widget("net-dns-expander").set_expanded(False)

        self.widget("net-ipv4-enable").set_active(True)
        self.widget("net-ipv4-network").set_text("192.168.100.0/24")
        self.widget("net-dhcpv4-enable").set_active(True)
        self.widget("net-dhcpv4-start").set_text("192.168.100.128")
        self.widget("net-dhcpv4-end").set_text("192.168.100.254")

        self.widget("net-ipv6-enable").set_active(False)
        self.widget("net-ipv6-network").set_text("")
        self.widget("net-dhcpv6-enable").set_active(False)
        self.widget("net-dhcpv6-start").set_text("")
        self.widget("net-dhcpv6-end").set_text("")

        # Populate physical forward devices
        self.widget("net-forward-device").set_active(0)
        self.widget("net-forward-mode").set_active(0)
        self.widget("net-forward-manual").set_text("")

        # Populate hostdev forward devices
        devprettynames = []
        ifnames = []
        for pcidev in self.conn.filter_nodedevs("pci"):
            if not pcidev.xmlobj.is_pci_sriov():
                continue
            devdesc = pcidev.pretty_name()
            for netdev in self.conn.filter_nodedevs("net"):
                if pcidev.xmlobj.name != netdev.xmlobj.parent:
                    continue
                ifname = netdev.xmlobj.interface
                devprettyname = "%s (%s)" % (ifname, devdesc)
                devprettynames.append(devprettyname)
                ifnames.append(ifname)
                break

        pf_model = self.widget("net-hostdevs").get_model()
        pf_model.clear()
        for devprettyname, ifname in zip(devprettynames, ifnames):
            pf_model.append([ifname, devprettyname])
        if len(pf_model) == 0:
            pf_model.append([None, _("No available device")])
        self.widget("net-hostdevs").set_active(0)

    ##################
    # UI get helpers #
    ##################

    def get_config_ipv4_enable(self):
        return (
            self.widget("net-ipv4-expander").is_visible()
            and self.widget("net-ipv4-enable").get_active()
        )

    def get_config_ipv6_enable(self):
        return (
            self.widget("net-ipv6-expander").is_visible()
            and self.widget("net-ipv6-enable").get_active()
        )

    def get_config_dhcpv4_enable(self):
        return self.widget("net-dhcpv4-enable").get_active()

    def get_config_dhcpv6_enable(self):
        return self.widget("net-dhcpv6-enable").get_active()

    def get_config_domain_name(self):
        widget = self.widget("net-domain-name")
        if not widget.is_visible():
            return None

        if self.widget("net-dns-use-netname").get_active():
            return self.widget("net-name").get_text()
        return widget.get_text()

    def _get_network_helper(self, widgetname):
        widget = self.widget(widgetname)
        if not widget.is_visible() or not widget.is_sensitive():
            return None
        return _make_ipaddr(widget.get_text())

    def get_config_ip4(self):
        return self._get_network_helper("net-ipv4-network")

    def get_config_dhcpv4_start(self):
        return self._get_network_helper("net-dhcpv4-start")

    def get_config_dhcpv4_end(self):
        return self._get_network_helper("net-dhcpv4-end")

    def get_config_ip6(self):
        return self._get_network_helper("net-ipv6-network")

    def get_config_dhcpv6_start(self):
        return self._get_network_helper("net-dhcpv6-start")

    def get_config_dhcpv6_end(self):
        return self._get_network_helper("net-dhcpv6-end")

    def _get_config_net_forward_dev(self):
        if not self.widget("net-forward-device").is_visible():
            return None

        manual = bool(uiutil.get_list_selection(self.widget("net-forward-device")))
        if not manual:
            return None
        return self.widget("net-forward-manual").get_text()

    #############
    # Listeners #
    #############

    def _net_forward_mode_changed_cb(self, src):
        mode = uiutil.get_list_selection(self.widget("net-forward-mode"))

        fw_visible = mode not in ["open", "isolated", "hostdev"]
        is_hostdev = mode in ["hostdev"]

        uiutil.set_grid_row_visible(self.widget("net-forward-device"), fw_visible)
        self._net_forward_device_changed_cb(self.widget("net-forward-device"))
        uiutil.set_grid_row_visible(self.widget("net-hostdevs"), is_hostdev)

        self.widget("net-ipv4-expander").set_visible(not is_hostdev)
        self.widget("net-ipv6-expander").set_visible(not is_hostdev)
        self.widget("net-dns-expander").set_visible(not is_hostdev)

    def _net_forward_device_changed_cb(self, src):
        manual = uiutil.get_list_selection(self.widget("net-forward-device"))
        if not src.is_visible():
            manual = False
        uiutil.set_grid_row_visible(self.widget("net-forward-manual"), manual)

    def _net_dns_use_toggled_cb(self, src):
        custom = self.widget("net-dns-use-custom").get_active()
        self.widget("net-domain-name").set_sensitive(custom)

    def _ipv4_toggled_cb(self, src):
        self.change_ipv4_enable()

    def _dhcpv4_toggled_cb(self, src):
        self.change_dhcpv4_enable()

    def _ipv6_toggled_cb(self, src):
        self.change_ipv6_enable()

    def _dhcpv6_toggled_cb(self, src):
        self.change_dhcpv6_enable()

    def change_ipv4_enable(self):
        enabled = self.widget("net-ipv4-enable").get_active()
        self.widget("net-ipv4-box").set_visible(enabled)

    def change_ipv6_enable(self):
        enabled = self.widget("net-ipv6-enable").get_active()
        self.widget("net-ipv6-box").set_visible(enabled)

    def change_dhcpv4_enable(self):
        enabled = self.get_config_dhcpv4_enable()
        start = self.widget("net-dhcpv4-start")
        end = self.widget("net-dhcpv4-end")
        uiutil.set_grid_row_visible(start, enabled)
        uiutil.set_grid_row_visible(end, enabled)

    def change_dhcpv6_enable(self):
        enabled = self.get_config_dhcpv6_enable()
        start = self.widget("net-dhcpv6-start")
        end = self.widget("net-dhcpv6-end")
        uiutil.set_grid_row_visible(start, enabled)
        uiutil.set_grid_row_visible(end, enabled)

    def _change_ipv4_network_cb(self, src):
        ip = self.get_config_ip4()

        # No IP specified or invalid IP
        if ip is None or ip.version != 4:
            return

        start = int(ip.num_addresses // 2)
        end = int(ip.num_addresses - 2)

        self.widget("net-dhcpv4-start").set_text(str(ip.network_address + start))
        self.widget("net-dhcpv4-end").set_text(str(ip.network_address + end))

    def _change_ipv6_network_cb(self, src):
        ip = self.get_config_ip6()

        if ip is None or ip.version != 6:
            return

        start = 256
        end = 512 - 1

        self.widget("net-dhcpv6-start").set_text(str(ip.network_address + start))
        self.widget("net-dhcpv6-end").set_text(str(ip.network_address + end))

    #########################
    # XML build and install #
    #########################

    def _validate(self, net):
        net.validate_generic_name(_("Network"), net.name)

        try:
            net.conn.networkLookupByName(net.name)
        except libvirt.libvirtError:
            return
        raise ValueError(_("Name '%s' already in use by another network.") % net.name)

    def _build_xmlobj_from_xmleditor(self):
        xml = self._xmleditor.get_xml()
        log.debug("Using XML from xmleditor:\n%s", xml)
        return Network(self.conn.get_backend(), parsexml=xml)

    def _build_xmlobj_from_ui(self):
        net = Network(self.conn.get_backend())

        net.name = self.widget("net-name").get_text()
        net.domain_name = self.get_config_domain_name()

        mode = uiutil.get_list_selection(self.widget("net-forward-mode"))
        dev = self._get_config_net_forward_dev()

        if mode == "isolated":
            mode = None

        net.forward.mode = mode
        net.forward.dev = dev

        if net.forward.mode == "hostdev":
            net.forward.managed = "yes"
            pfobj = net.forward.pf.add_new()
            pfobj.dev = uiutil.get_list_selection(self.widget("net-hostdevs"))
            return net

        if self.get_config_ipv4_enable():
            ip = self.get_config_ip4()
            ipobj = net.ips.add_new()
            if ip:
                ipobj.address = str(ip.network_address + 1)
                ipobj.netmask = str(ip.netmask)

            if self.get_config_dhcpv4_enable():
                dhcpobj = ipobj.ranges.add_new()
                start = self.get_config_dhcpv4_start()
                end = self.get_config_dhcpv4_end()
                if start:
                    dhcpobj.start = str(start.network_address)
                if end:
                    dhcpobj.end = str(end.network_address)

        if self.get_config_ipv6_enable():
            ip = self.get_config_ip6()
            ipobj = net.ips.add_new()
            ipobj.family = "ipv6"
            if ip:
                ipobj.address = str(ip.network_address + 1)
                ipobj.prefix = str(ip.prefixlen)

            if self.get_config_dhcpv6_enable():
                dhcpobj = ipobj.ranges.add_new()
                start = self.get_config_dhcpv6_start()
                end = self.get_config_dhcpv6_end()
                if start:
                    dhcpobj.start = str(start.network_address)
                if end:
                    dhcpobj.end = str(end.network_address)

        return net

    def _build_xmlobj(self, check_xmleditor):
        try:
            xmlobj = self._build_xmlobj_from_ui()
            if check_xmleditor and self._xmleditor.is_xml_selected():
                xmlobj = self._build_xmlobj_from_xmleditor()
            return xmlobj
        except Exception as e:
            self.err.show_err(_("Error building XML: %s") % str(e))

    def _finish_cb(self, error, details):
        self.reset_finish_cursor()

        if error:
            error = _("Error creating virtual network: %s") % str(error)
            self.err.show_err(error, details=details)
        else:
            self.conn.schedule_priority_tick(pollnet=True)
            self.close()

    def _async_net_create(self, asyncjob, net):
        ignore = asyncjob
        xml = net.get_xml()
        log.debug("Creating virtual network '%s' with xml:\n%s", net.name, xml)

        netobj = self.conn.get_backend().networkDefineXML(xml)
        try:
            netobj.create()
            netobj.setAutostart(True)
        except Exception:  # pragma: no cover
            netobj.undefine()
            raise

    def finish(self, ignore):
        net = self._build_xmlobj(check_xmleditor=True)
        if not net:
            return

        try:
            self._validate(net)
        except Exception as e:
            return self.err.show_err(_("Error validating network: %s") % e)

        self.set_finish_cursor()
        progWin = vmmAsyncJob(
            self._async_net_create,
            [net],
            self._finish_cb,
            [],
            _("Creating virtual network..."),
            _("Creating the virtual network may take a while..."),
            self.topwin,
        )
        progWin.run()

    ################
    # UI listeners #
    ################

    def _xmleditor_xml_requested_cb(self, src):
        xmlobj = self._build_xmlobj(check_xmleditor=False)
        self._xmleditor.set_xml(xmlobj and xmlobj.get_xml() or "")



# Copyright (C) 2006, 2014 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import argparse
import os
import signal
import sys
import traceback

import gi

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
gi.require_version("LibvirtGLib", "1.0")
from gi.repository import LibvirtGLib

from virtinst import BuildConfig
from virtinst import cli
from virtinst import log

from .lib.testmock import CLITestOptionsClass

# pygobject commit that I believe universally changed bool arguments
# to handle passed None. Without this, None usage keeps slipping into
# the code, so add the requirement.
# https://github.com/GNOME/pygobject/commit/6c69fc7b582ec1fd3faef4de3fade9a0cb7f8c05
_PYGOBJECT_VERSION = "3.31.3"
try:
    gi.check_version(_PYGOBJECT_VERSION)
except (ValueError, AttributeError):  # pragma: no cover
    print("pygobject3 %s or later is required." % _PYGOBJECT_VERSION)
    sys.exit(1)


def _show_startup_error(msg, details):
    log.debug("Error starting virt-manager: %s\n%s", msg, details, exc_info=True)
    from .error import vmmErrorDialog

    err = vmmErrorDialog.get_instance()
    title = _("Error starting Virtual Machine Manager")
    errmsg = _("Error starting Virtual Machine Manager: %(error)s") % {"error": msg}
    err.show_err(errmsg, details=details, title=title, modal=True, debug=False)


def _import_gtk(leftovers):
    # The never ending fork+gsettings problems now require us to
    # import Gtk _after_ the fork. This creates a funny race, since we
    # need to parse the command line arguments to know if we need to
    # fork, but need to import Gtk before cli processing so it can
    # handle --g-fatal-args. We strip out our flags first and pass the
    # left overs to gtk
    origargv = sys.argv
    try:
        sys.argv = origargv[:1] + leftovers[:]
        from gi.repository import Gtk

        leftovers = sys.argv[1:]

        if Gtk.check_version(3, 22, 0):  # pragma: no cover
            print("gtk3 3.22.0 or later is required.")
            sys.exit(1)

        # This will error if Gtk wasn't correctly initialized
        Gtk.init()
        globals()["Gtk"] = Gtk

        # This ensures we can init gsettings correctly
        from . import config

        ignore = config
    finally:
        sys.argv = origargv

    return leftovers


def _setup_gsettings_path(schemadir):
    """
    If running from the virt-manager.git srcdir, compile our gsettings
    schema and use it directly
    """
    import subprocess
    import shutil

    exe = shutil.which("glib-compile-schemas")
    if not exe:  # pragma: no cover
        raise RuntimeError("You must install glib-compile-schemas to run virt-manager from git.")
    subprocess.check_call([exe, "--strict", schemadir])


def drop_tty():
    # We fork and setsid so that we drop the controlling
    # tty. This prevents libvirt's SSH tunnels from prompting
    # for user input if SSH keys/agent aren't configured.
    if os.fork() != 0:
        # pylint: disable=protected-access
        os._exit(0)  # pragma: no cover

    os.setsid()


def drop_stdio():
    # This is part of the fork process described in drop_tty()
    for fd in range(0, 2):
        try:
            os.close(fd)
        except OSError:  # pragma: no cover
            pass

    os.open(os.devnull, os.O_RDWR)
    os.dup2(0, 1)
    os.dup2(0, 2)


def do_we_fork(options):
    if options.debug or options.no_fork:
        return False
    if options.fork:
        return True

    key = "VIRT_MANAGER_DEFAULT_FORK"
    val = os.environ.get(key, None)
    if val == "yes":
        log.debug("%s=%s, defaulting to --fork", key, val)
        return True
    if val == "no":  # pragma: no cover
        log.debug("%s=%s, defaulting to --no-fork", key, val)
        return False
    if val:  # pragma: no cover
        log.warning("Unknown %s=%s, expected 'yes' or 'no'", key, val)

    # Default is `--no-fork`
    return False


def parse_commandline():
    epilog = "Also accepts standard GTK arguments like --g-fatal-warnings"
    parser = argparse.ArgumentParser(usage="virt-manager [options]", epilog=epilog)
    parser.add_argument("--version", action="version", version=BuildConfig.version)
    parser.set_defaults(domain=None)

    # Trace every libvirt API call to debug output
    parser.add_argument("--trace-libvirt", choices=["all", "mainloop"], help=argparse.SUPPRESS)

    # comma separated string of options to tweak app behavior,
    # for manual and automated testing config
    parser.add_argument("--test-options", action="append", default=[], help=argparse.SUPPRESS)

    parser.add_argument(
        "-c", "--connect", dest="uri", help="Connect to hypervisor at URI", metavar="URI"
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Print debug output to stdout (implies --no-fork)",
        default=False,
    )
    parser.add_argument(
        "--no-fork", action="store_true", help="Don't fork into background on startup"
    )
    parser.add_argument(
        "--fork",
        action="store_true",
        help="Force fork into background on startup (this is the default)",
    )

    parser.add_argument("--show-domain-creator", action="store_true", help="Show 'New VM' wizard")
    parser.add_argument(
        "--show-domain-editor", metavar="NAME|ID|UUID", help="Show domain details window"
    )
    parser.add_argument(
        "--show-domain-performance", metavar="NAME|ID|UUID", help="Show domain performance window"
    )
    parser.add_argument(
        "--show-domain-console", metavar="NAME|ID|UUID", help="Show domain graphical console window"
    )
    parser.add_argument(
        "--show-domain-delete", metavar="NAME|ID|UUID", help="Show domain delete window"
    )
    parser.add_argument(
        "--show-host-summary", action="store_true", help="Show connection details window"
    )
    parser.add_argument(
        "--show-systray", action="store_true", help="Launch virt-manager only in system tray"
    )

    return parser.parse_known_args()


def main():
    (options, leftovers) = parse_commandline()

    cli.setupLogging("virt-manager", options.debug, False, False)

    log.debug("virt-manager version: %s", BuildConfig.version)
    log.debug("virtManager import: %s", os.path.dirname(__file__))

    if BuildConfig.running_from_srcdir:
        _setup_gsettings_path(BuildConfig.gsettings_dir)

    if options.trace_libvirt:
        log.debug("Libvirt tracing requested")
        from .lib import module_trace
        import libvirt

        module_trace.wrap_module(
            libvirt, mainloop=(options.trace_libvirt == "mainloop"), regex=None
        )

    CLITestOptions = CLITestOptionsClass(options.test_options)

    # With F27 gnome+wayland we need to set these before GTK import
    os.environ["GSETTINGS_SCHEMA_DIR"] = BuildConfig.gsettings_dir

    # Force SSH to use askpass if a password is required,
    # rather than possibly prompting on a terminal the user isn't looking at.
    os.environ.setdefault("SSH_ASKPASS_REQUIRE", "force")
    log.debug("Using SSH_ASKPASS_REQUIRE=%s", os.environ["SSH_ASKPASS_REQUIRE"])

    # Now we've got basic environment up & running we can fork
    do_fork = do_we_fork(options)
    if do_fork:
        drop_tty()

    leftovers = _import_gtk(leftovers)
    Gtk = globals()["Gtk"]

    # Do this after the Gtk import so the user has a chance of seeing any error
    if do_fork:
        drop_stdio()

    if leftovers:
        raise RuntimeError("Unhandled command line options '%s'" % leftovers)

    log.debug(
        "PyGObject version: %d.%d.%d", gi.version_info[0], gi.version_info[1], gi.version_info[2]
    )
    log.debug(
        "GTK version: %d.%d.%d",
        Gtk.get_major_version(),
        Gtk.get_minor_version(),
        Gtk.get_micro_version(),
    )

    # Prime the vmmConfig cache
    from . import config

    config.vmmConfig.get_instance(BuildConfig, CLITestOptions)

    # Add our icon dir to icon theme
    icon_theme = Gtk.IconTheme.get_default()
    icon_theme.prepend_search_path(BuildConfig.icon_dir)

    from .engine import vmmEngine

    Gtk.Window.set_default_icon_name("virt-manager")

    show_window = None
    domain = None
    if options.show_domain_creator:
        show_window = vmmEngine.CLI_SHOW_DOMAIN_CREATOR
    elif options.show_host_summary:
        show_window = vmmEngine.CLI_SHOW_HOST_SUMMARY
    elif options.show_domain_editor:
        show_window = vmmEngine.CLI_SHOW_DOMAIN_EDITOR
        domain = options.show_domain_editor
    elif options.show_domain_performance:
        show_window = vmmEngine.CLI_SHOW_DOMAIN_PERFORMANCE
        domain = options.show_domain_performance
    elif options.show_domain_console:
        show_window = vmmEngine.CLI_SHOW_DOMAIN_CONSOLE
        domain = options.show_domain_console
    elif options.show_domain_delete:
        show_window = vmmEngine.CLI_SHOW_DOMAIN_DELETE
        domain = options.show_domain_delete
    elif options.show_systray:
        show_window = vmmEngine.CLI_SHOW_SYSTEM_TRAY

    if (
        show_window and show_window != vmmEngine.CLI_SHOW_SYSTEM_TRAY and options.uri is None
    ):  # pragma: no cover
        raise RuntimeError("can't use --show-* options without --connect (except --show-systray)")

    skip_autostart = False
    if show_window and show_window != vmmEngine.CLI_SHOW_SYSTEM_TRAY:
        skip_autostart = True

    # Hook libvirt events into glib main loop
    LibvirtGLib.init(None)
    LibvirtGLib.event_register()

    engine = vmmEngine.get_instance()

    # Actually exit when we receive ctrl-c
    from gi.repository import GLib

    def _sigint_handler(user_data):
        ignore = user_data
        log.debug("Received KeyboardInterrupt. Exiting application.")
        engine.exit_app()

    GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGINT, _sigint_handler, None)

    engine.start(options.uri, show_window, domain, skip_autostart)


def runcli():
    try:
        main()
    except KeyboardInterrupt:  # pragma: no cover
        log.debug("Received KeyboardInterrupt. Exiting application.")
    except Exception as run_e:
        if "Gtk" not in globals():
            raise  # pragma: no cover
        _show_startup_error(str(run_e), "".join(traceback.format_exc()))



#
# Copyright 2009, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os
import uuid

from .xmlbuilder import XMLBuilder, XMLProperty, XMLChildProperty


def _compare_int(nodedev_val, hostdev_val):
    def _intify(val):
        try:
            if "0x" in str(val):
                return int(val or "0x00", 16)
            else:
                return int(val)
        except Exception:
            return -1

    nodedev_val = _intify(nodedev_val)
    hostdev_val = _intify(hostdev_val)
    return nodedev_val == hostdev_val or hostdev_val == -1


def _compare_uuid(nodedev_val, hostdev_val):
    try:
        nodedev_val = uuid.UUID(nodedev_val)
        hostdev_val = uuid.UUID(hostdev_val)
    except Exception:  # pragma: no cover
        return -1

    return nodedev_val == hostdev_val


class DevNode(XMLBuilder):
    XML_NAME = "devnode"

    node_type = XMLProperty("./@type")
    path = XMLProperty(".")


class NodeDevice(XMLBuilder):
    CAPABILITY_TYPE_NET = "net"
    CAPABILITY_TYPE_PCI = "pci"
    CAPABILITY_TYPE_USBDEV = "usb_device"
    CAPABILITY_TYPE_STORAGE = "storage"
    CAPABILITY_TYPE_SCSIBUS = "scsi_host"
    CAPABILITY_TYPE_SCSIDEV = "scsi"
    CAPABILITY_TYPE_DRM = "drm"
    CAPABILITY_TYPE_MDEV = "mdev"

    @staticmethod
    def lookupNodedevByName(conn, name):
        """
        Search the nodedev list cache for a matching name, and return the
        result.

        :param conn: libvirt.virConnect instance to perform the lookup on
        :param conn: nodedev name
        :returns: NodeDevice instance
        """
        for nodedev in conn.fetch_all_nodedevs():
            if nodedev.name == name:
                return nodedev

    XML_NAME = "device"

    # Libvirt can generate bogus 'system' XML:
    # https://bugzilla.redhat.com/show_bug.cgi?id=1184131
    _XML_SANITIZE = True

    name = XMLProperty("./name")
    parent = XMLProperty("./parent")
    device_type = XMLProperty("./capability/@type")

    def get_mdev_uuid(self):
        # libvirt 7.3.0 added a <uuid> element to the nodedev xml for mdev
        # types. For older versions, we unfortunately have to parse the nodedev
        # name, which uses the format "mdev_$UUID_WITH_UNDERSCORES"
        if self.uuid is not None:
            return self.uuid

        return self.name[5:].replace("_", "-")

    def compare_to_hostdev(self, hostdev):
        if self.device_type == "pci":
            if hostdev.type != "pci":
                return False

            return (
                _compare_int(self.domain, hostdev.domain)
                and _compare_int(self.bus, hostdev.bus)
                and _compare_int(self.slot, hostdev.slot)
                and _compare_int(self.function, hostdev.function)
            )

        if self.device_type == "usb_device":
            if hostdev.type != "usb":
                return False

            return (
                _compare_int(self.product_id, hostdev.product)
                and _compare_int(self.vendor_id, hostdev.vendor)
                and _compare_int(self.bus, hostdev.bus)
                and _compare_int(self.device, hostdev.device)
            )

        if self.device_type == "mdev":
            if hostdev.type != "mdev":
                return False

            return _compare_uuid(self.get_mdev_uuid(), hostdev.uuid)

        return False

    ########################
    # XML helper functions #
    ########################

    def is_pci_sriov(self):
        return self._capability_type == "virt_functions"

    def is_pci_bridge(self):
        return self._capability_type == "pci-bridge"

    def is_usb_linux_root_hub(self):
        return self.vendor_id == "0x1d6b" and self.product_id in ["0x0001", "0x0002", "0x0003"]

    def is_drm_render(self):
        return self.device_type == "drm" and self.drm_type == "render"

    ##################
    # XML properties #
    ##################

    # type='net' options
    interface = XMLProperty("./capability/interface")

    # type='pci' options
    domain = XMLProperty("./capability/domain")
    bus = XMLProperty("./capability/bus")
    slot = XMLProperty("./capability/slot")
    function = XMLProperty("./capability/function")
    product_name = XMLProperty("./capability/product")
    vendor_name = XMLProperty("./capability/vendor")
    _capability_type = XMLProperty("./capability/capability/@type")

    # type='usb' options
    device = XMLProperty("./capability/device")
    product_id = XMLProperty("./capability/product/@id")
    vendor_id = XMLProperty("./capability/vendor/@id")

    # type='scsi' options
    host = XMLProperty("./capability/host")
    target = XMLProperty("./capability/target")
    lun = XMLProperty("./capability/lun")

    # type='storage' options
    block = XMLProperty("./capability/block")
    drive_type = XMLProperty("./capability/drive_type")

    media_label = XMLProperty("./capability/capability[@type='removable']/media_label")
    media_available = XMLProperty(
        "./capability/capability[@type='removable']/media_available", is_int=True
    )

    # type='drm' options
    drm_type = XMLProperty("./capability/type")
    devnodes = XMLChildProperty(DevNode)

    def get_devnode(self, parent="by-path"):
        for d in self.devnodes:
            paths = d.path.split(os.sep)
            if len(paths) > 2 and paths[-2] == parent:
                return d
        if len(self.devnodes) > 0:
            return self.devnodes[0]

    # type='mdev' options
    type_id = XMLProperty("./capability/type/@id")
    uuid = XMLProperty("./capability/uuid")



#
# Copyright 2010, 2013 Red Hat, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceWatchdog(Device):
    XML_NAME = "watchdog"

    MODEL_I6300 = "i6300esb"
    MODEL_IB700 = "ib700"
    MODEL_DIAG288 = "diag288"
    MODELS = [MODEL_I6300, MODEL_IB700, MODEL_DIAG288]

    ACTION_SHUTDOWN = "shutdown"
    ACTION_RESET = "reset"
    ACTION_POWEROFF = "poweroff"
    ACTION_PAUSE = "pause"
    ACTION_NONE = "none"
    ACTION_DUMP = "dump"
    ACTIONS = [
        ACTION_RESET,
        ACTION_SHUTDOWN,
        ACTION_POWEROFF,
        ACTION_PAUSE,
        ACTION_DUMP,
        ACTION_NONE,
    ]

    _XML_PROP_ORDER = ["model", "action"]
    model = XMLProperty("./@model")
    action = XMLProperty("./@action")

    ##################
    # Default config #
    ##################

    def set_defaults(self, _guest):
        if not self.model:
            self.model = self.MODEL_I6300
        if not self.action:
            self.action = self.ACTION_RESET



# Copyright (C) 2006-2008, 2015 Red Hat, Inc.
# Copyright (C) 2006 Daniel P. Berrange <berrange@redhat.com>
# Copyright (C) 2010 Marc-Andre Lureau <marcandre.lureau@redhat.com>
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

import os

from gi.repository import Gdk
from gi.repository import GObject

import gi

gi.require_version("GtkVnc", "2.0")
from gi.repository import GtkVnc

try:
    SPICE_GTK_IMPORT_ERROR = None
    if "VIRTINST_TEST_SUITE_FAKE_NO_SPICE" in os.environ:
        raise ImportError("test suite faking no spice")

    gi.require_version("SpiceClientGtk", "3.0")
    from gi.repository import SpiceClientGtk
    from gi.repository import SpiceClientGLib
except (ValueError, ImportError) as _SPICE_GTK_IMPORT_ERROR:
    SPICE_GTK_IMPORT_ERROR = str(_SPICE_GTK_IMPORT_ERROR)

from virtinst import log

from .sshtunnels import SSHTunnels
from ..baseclass import vmmGObject


##################################
# VNC/Spice abstraction handling #
##################################

_GTKVNC_SUPPORT_CACHE = {}


def _gtkvnc_check_display_support(funcname):
    if funcname not in _GTKVNC_SUPPORT_CACHE:
        val = hasattr(GtkVnc.Display, funcname)
        log.debug("GtkVnc.Display %s support=%s", funcname, val)
        _GTKVNC_SUPPORT_CACHE[funcname] = val
    return _GTKVNC_SUPPORT_CACHE[funcname]


def _gtkvnc_supports_resizeguest():
    return _gtkvnc_check_display_support("set_allow_resize")


class Viewer(vmmGObject):
    """
    Base class for viewer abstraction
    """

    __gsignals__ = {
        "add-display-widget": (vmmGObject.RUN_FIRST, None, [object]),
        "size-allocate": (vmmGObject.RUN_FIRST, None, [object]),
        "desktop-resolution-changed": (vmmGObject.RUN_FIRST, None, []),
        "pointer-grab": (vmmGObject.RUN_FIRST, None, []),
        "pointer-ungrab": (vmmGObject.RUN_FIRST, None, []),
        "keyboard-grab": (vmmGObject.RUN_FIRST, None, []),
        "keyboard-ungrab": (vmmGObject.RUN_FIRST, None, []),
        "connected": (vmmGObject.RUN_FIRST, None, []),
        "disconnected": (vmmGObject.RUN_FIRST, None, [str, str]),
        "auth-error": (vmmGObject.RUN_FIRST, None, [str, bool]),
        "need-auth": (vmmGObject.RUN_FIRST, None, [bool, bool]),
        "agent-connected": (vmmGObject.RUN_FIRST, None, []),
        "usb-redirect-error": (vmmGObject.RUN_FIRST, None, [str]),
    }

    def __init__(self, vm, ginfo):
        vmmGObject.__init__(self)
        self._display = None
        self._vm = vm
        self._ginfo = ginfo
        self._tunnels = SSHTunnels(self._ginfo)
        self._keyboard_grab = False
        self._desktop_resolution = (0, 0)

        self.add_gsettings_handle(self.config.on_keys_combination_changed(self._refresh_grab_keys))

    def _cleanup(self):
        self._close()

        if self._display:
            self._display.destroy()
        self._display = None
        self._vm = None

        self._tunnels.close_all()

    ########################
    # Internal helper APIs #
    ########################

    def _make_signal_proxy(self, new_signal):
        """
        Helper for redirecting a signal from self._display out
        through the viewer
        """

        def _proxy_signal(src, *args, **kwargs):
            ignore = src
            self.emit(new_signal, *args, **kwargs)

        return _proxy_signal

    def _set_display(self, display):
        self._display = display
        self._refresh_grab_keys()

        self._display.connect("size-allocate", self._make_signal_proxy("size-allocate"))

        self.emit("add-display-widget", self._display)
        self._display.realize()

        self._connect_display_signals()

        self._display.show()

    #########################
    # Generic internal APIs #
    #########################

    def _grab_focus(self):
        if self._display:
            self._display.grab_focus()

    def _get_pixbuf(self):
        return self._display.get_pixbuf()

    def _get_fd_for_open(self):
        if self._ginfo.need_tunnel():
            return self._tunnels.open_new()

        if self._vm.conn.is_remote():
            # OpenGraphics only works for local libvirtd connections
            return None

        if self._ginfo.gtlsport and not self._ginfo.gport:
            # This makes spice loop requesting an fd. Disable until spice is
            # fixed: https://bugzilla.redhat.com/show_bug.cgi?id=1334071
            return None  # pragma: no cover

        if not self._vm.conn.support.domain_open_graphics():
            return None

        return self._vm.open_graphics_fd(self._ginfo.gidx)

    def _open(self):
        if self._ginfo.bad_config():
            raise RuntimeError(self._ginfo.bad_config())

        fd = self._get_fd_for_open()
        if fd is not None:
            self._open_fd(fd)
        else:
            self._open_host()

    def _get_grab_keys(self):
        return self._display.get_grab_keys().as_string()

    def _refresh_grab_keys(self):
        if not self._display:
            return  # pragma: no cover

        try:
            keys = self.config.get_keys_combination()
            if not keys:
                return  # pragma: no cover

            try:
                keys = [int(k) for k in keys.split(",")]
            except Exception:  # pragma: no cover
                log.debug("Error in grab_keys configuration in Gsettings", exc_info=True)
                return

            self._set_grab_keys(keys)
        except Exception as e:  # pragma: no cover
            log.debug("Error when getting the grab keys combination: %s", str(e))

    def _emit_disconnected(self, errdetails=None):
        ssherr = self._tunnels.get_err_output()
        self.emit("disconnected", errdetails, ssherr)

    def _set_desktop_resolution(self, w, h):
        if w == 0 or h == 0:
            return  # pragma: no cover

        origres = self._desktop_resolution
        newres = (w, h)
        if origres == newres:
            return  # pragma: no cover

        log.debug("vm=%s resolution changed: orig=%s new=%s", self._vm.get_name(), origres, newres)
        self._desktop_resolution = newres
        self.emit("desktop-resolution-changed")

    #######################################################
    # Internal API that will be overwritten by subclasses #
    #######################################################

    def _close(self):
        raise NotImplementedError()

    def _is_open(self):
        raise NotImplementedError()

    def _connect_display_signals(self):
        raise NotImplementedError()

    def _set_username(self, cred):
        raise NotImplementedError()

    def _set_password(self, cred):
        raise NotImplementedError()

    def _send_keys(self, keys):
        raise NotImplementedError()

    def _set_grab_keys(self, keys):
        raise NotImplementedError()

    def _open_host(self):
        raise NotImplementedError()

    def _open_fd(self, fd):
        raise NotImplementedError()

    def _get_scaling(self):
        raise NotImplementedError()

    def _set_scaling(self, scaling):
        raise NotImplementedError()

    def _set_resizeguest(self, val):
        raise NotImplementedError()

    def _get_resizeguest(self):
        raise NotImplementedError()

    def _get_resizeguest_warning(self):
        raise NotImplementedError()

    def _get_usb_widget(self):
        raise NotImplementedError()

    def _has_usb_redirection(self):
        raise NotImplementedError()

    def _get_preferred_size(self):
        raise NotImplementedError()

    ####################################
    # APIs accessed by vmmConsolePages #
    ####################################

    def console_is_open(self):
        return self._is_open()

    def console_grab_focus(self):
        return self._grab_focus()

    def console_has_keyboard_grab(self):
        return bool(self._display and self._keyboard_grab)

    def console_get_pixbuf(self):
        return self._get_pixbuf()

    def console_open(self):
        return self._open()

    def console_set_password(self, val):
        return self._set_password(val)

    def console_set_username(self, val):
        return self._set_username(val)

    def console_send_keys(self, keys):
        return self._send_keys(keys)

    def console_get_grab_keys(self):
        return self._get_grab_keys()

    def console_get_preferred_size(self):
        return self._get_preferred_size()

    def console_set_scaling(self, val):
        return self._set_scaling(val)

    def console_set_resizeguest(self, val):
        return self._set_resizeguest(val)

    def console_get_resizeguest_warning(self):
        return self._get_resizeguest_warning()

    def console_get_usb_widget(self):
        return self._get_usb_widget()

    def console_has_usb_redirection(self):
        return self._has_usb_redirection()

    def console_remove_display_from_widget(self, widget):
        if self._display and self._display in widget.get_children():
            widget.remove(self._display)


####################
# VNC viewer class #
####################


class VNCViewer(Viewer):
    viewer_type = "vnc"

    ###################
    # Private helpers #
    ###################

    def _connect_display_signals(self):
        self._display.connect("vnc-pointer-grab", self._make_signal_proxy("pointer-grab"))
        self._display.connect("vnc-pointer-ungrab", self._make_signal_proxy("pointer-ungrab"))
        self._display.connect("vnc-keyboard-grab", self._keyboard_grab_cb)
        self._display.connect("vnc-keyboard-ungrab", self._keyboard_ungrab_cb)

        self._display.connect("vnc-auth-credential", self._auth_credential)
        self._display.connect("vnc-auth-failure", self._auth_failure_cb)
        self._display.connect("vnc-initialized", self._connected_cb)
        self._display.connect("vnc-disconnected", self._disconnected_cb)
        self._display.connect("vnc-desktop-resize", self._desktop_resize_cb)

    def _init_display(self):
        display = GtkVnc.Display()

        display.set_pointer_grab(True)

        if _gtkvnc_check_display_support("set_keep_aspect_ratio"):
            display.set_keep_aspect_ratio(True)

        self._set_display(display)

    def _keyboard_grab_cb(self, src):
        self._keyboard_grab = True
        self.emit("keyboard-grab")

    def _keyboard_ungrab_cb(self, src):
        self._keyboard_grab = False
        self.emit("keyboard-ungrab")

    def _connected_cb(self, ignore):
        self._tunnels.unlock()
        self.emit("connected")

    def _disconnected_cb(self, ignore):
        self._tunnels.unlock()
        self._emit_disconnected()

    def _desktop_resize_cb(self, src, w, h):
        self._set_desktop_resolution(w, h)

    def _auth_failure_cb(self, ignore, msg):
        log.debug("VNC auth failure. msg=%s", msg)
        self.emit("auth-error", msg, True)

    def _auth_credential(self, src_ignore, credList):
        values = []
        for idx in range(int(credList.n_values)):
            values.append(credList.get_nth(idx))

        if self.config.CLITestOptions.fake_vnc_username:
            values.append(GtkVnc.DisplayCredential.USERNAME)

        withUsername = False
        withPassword = False
        for cred in values:
            log.debug("Got credential request %s", cred)
            if cred == GtkVnc.DisplayCredential.PASSWORD:
                withPassword = True
            elif cred == GtkVnc.DisplayCredential.USERNAME:
                withUsername = True
            elif cred == GtkVnc.DisplayCredential.CLIENTNAME:  # pragma: no cover
                self._display.set_credential(cred, "libvirt-vnc")
            else:  # pragma: no cover
                errmsg = _(
                    "Unable to provide requested credentials to the VNC server.\n"
                    "The credential type %s is not supported"
                ) % str(cred.value_name)
                self.emit("auth-error", errmsg, True)
                return

        if withUsername or withPassword:
            self.emit("need-auth", withPassword, withUsername)

    def _sync_force_size(self):
        force_size = not self._get_scaling() and not self._get_resizeguest()
        self._display.set_force_size(force_size)

    ###############################
    # Private API implementations #
    ###############################

    def _close(self):
        self._display.close()

    def _is_open(self):
        return self._display.is_open()

    def _get_scaling(self):
        if self._display:
            return self._display.get_scaling()

    def _set_scaling(self, scaling):
        if self._display:
            self._display.set_scaling(scaling)
            self._sync_force_size()

    def _set_grab_keys(self, keys):
        seq = GtkVnc.GrabSequence.new(keys)
        self._display.set_grab_keys(seq)

    def _send_keys(self, keys):
        return self._display.send_keys([Gdk.keyval_from_name(k) for k in keys])

    def _set_username(self, cred):
        self._display.set_credential(GtkVnc.DisplayCredential.USERNAME, cred)

    def _set_password(self, cred):
        self._display.set_credential(GtkVnc.DisplayCredential.PASSWORD, cred)

    def _set_resizeguest(self, val):
        if _gtkvnc_supports_resizeguest():
            self._display.set_allow_resize(val)  # pylint: disable=no-member
            self._sync_force_size()

    def _get_resizeguest(self):
        if _gtkvnc_supports_resizeguest():
            return self._display.get_allow_resize()  # pylint: disable=no-member
        return False  # pragma: no cover

    def _get_resizeguest_warning(self):
        if not _gtkvnc_supports_resizeguest():
            return _("GTK-VNC viewer is too old")  # pragma: no cover

    def _get_usb_widget(self):
        return None  # pragma: no cover

    def _has_usb_redirection(self):
        return False

    def _get_preferred_size(self):
        return self._desktop_resolution

    #######################
    # Connection routines #
    #######################

    def _open(self):
        self._init_display()
        return Viewer._open(self)

    def _open_host(self):
        host, port, ignore = self._ginfo.get_conn_host()
        log.debug("VNC connecting to host=%s port=%s", host, port)
        self._display.open_host(host, port)

    def _open_fd(self, fd):
        self._display.open_fd(fd)


######################
# Spice viewer class #
######################


class _SignalTracker:
    # Helper class to more conveniently connect and disconnect signals
    # from spice objects. This ensures we don't leave circular references
    # at object destroy time
    def __init__(self):
        self._sigmap = {}

    def _add_hid(self, obj, hid):
        if obj not in self._sigmap:
            self._sigmap[obj] = []
        self._sigmap[obj].append(hid)

    def connect(self, obj, name, handler, *args):
        hid = GObject.GObject.connect(obj, name, handler, *args)
        self._add_hid(obj, hid)

    def connect_after(self, obj, name, handler, *args):
        hid = GObject.GObject.connect_after(obj, name, handler, *args)
        self._add_hid(obj, hid)

    def disconnect_obj_signals(self, obj):
        for hid in self._sigmap.get(obj, []):
            GObject.GObject.disconnect(obj, hid)


_SIGS = _SignalTracker()


class SpiceViewer(Viewer):
    viewer_type = "spice"

    def __init__(self, *args, **kwargs):
        Viewer.__init__(self, *args, **kwargs)
        self._spice_session = None
        self._display = None
        self._audio = None
        self._main_channel = None
        self._usbdev_manager = None
        self._channels = set()

    ###################
    # Private helpers #
    ###################

    def _mouse_grab_cb(self, src, grab):
        if grab:
            self.emit("pointer-grab")
        else:
            self.emit("pointer-ungrab")

    def _keyboard_grab_cb(self, src, grab):
        self._keyboard_grab = grab
        if grab:
            self.emit("keyboard-grab")
        else:
            self.emit("keyboard-ungrab")

    def _connect_display_signals(self):
        self._display.connect("mouse-grab", self._mouse_grab_cb)
        self._display.connect("keyboard-grab", self._keyboard_grab_cb)

    def _create_spice_session(self):
        self._spice_session = SpiceClientGLib.Session()
        SpiceClientGLib.set_session_option(self._spice_session)
        gtk_session = SpiceClientGtk.GtkSession.get(self._spice_session)
        gtk_session.set_property("auto-clipboard", True)

        _SIGS.connect(self._spice_session, "channel-new", self._channel_new_cb)

        # Distros might have usb redirection compiled out, like OpenBSD
        # https://bugzilla.redhat.com/show_bug.cgi?id=1348479
        try:
            self._usbdev_manager = SpiceClientGLib.UsbDeviceManager.get(self._spice_session)
            _SIGS.connect(self._usbdev_manager, "auto-connect-failed", self._usbdev_redirect_error)
            _SIGS.connect(self._usbdev_manager, "device-error", self._usbdev_redirect_error)

            autoredir = self.config.get_auto_usbredir()
            if autoredir:
                gtk_session.set_property("auto-usbredir", True)
        except Exception:  # pragma: no cover
            self._usbdev_manager = None
            log.debug("Error initializing spice usb device manager", exc_info=True)

    #####################
    # Channel listeners #
    #####################

    def _main_channel_event_cb(self, channel, event):
        self._tunnels.unlock()

        if event == SpiceClientGLib.ChannelEvent.CLOSED:
            self._emit_disconnected()
        elif event == SpiceClientGLib.ChannelEvent.ERROR_AUTH:
            if not self._spice_session.get_property("password"):
                log.debug(
                    "Spice channel received ERROR_AUTH, but no "
                    "password set, assuming it wants credentials."
                )
                self.emit("need-auth", True, False)
            else:
                log.debug(
                    "Spice channel received ERROR_AUTH, but a "
                    "password is already set. Assuming authentication failed."
                )
                self.emit("auth-error", channel.get_error().message, False)
        elif "ERROR" in str(event):
            # SpiceClientGLib.ChannelEvent.ERROR_CONNECT
            # SpiceClientGLib.ChannelEvent.ERROR_IO
            # SpiceClientGLib.ChannelEvent.ERROR_LINK
            # SpiceClientGLib.ChannelEvent.ERROR_TLS
            error = None
            if channel.get_error():
                error = channel.get_error().message
            log.debug("Spice channel event=%s message=%s", event, error)

            msg = _("Encountered SPICE %(error-name)s") % {"error-name": event.value_nick}
            if error:
                msg += ": %s" % error
            self._emit_disconnected(msg)

    def _fd_channel_event_cb(self, channel, event):
        # When we see any event from the channel, release the
        # associated tunnel lock
        channel.disconnect_by_func(self._fd_channel_event_cb)
        self._tunnels.unlock()

    def _channel_open_fd_request(self, channel, tls_ignore):
        if not self._tunnels:
            # Can happen if we close the details window and clear self._tunnels
            # while initially connecting to spice and channel FD requests
            # are still rolling in
            return  # pragma: no cover

        log.debug("Requesting fd for channel: %s", channel)
        channel.connect_after("channel-event", self._fd_channel_event_cb)

        fd = self._get_fd_for_open()
        channel.open_fd(fd)

    def _channel_new_cb(self, session, channel):
        self._channels.add(channel)
        _SIGS.connect(channel, "open-fd", self._channel_open_fd_request)

        if isinstance(channel, SpiceClientGLib.MainChannel) and not self._main_channel:
            self._main_channel = channel
            _SIGS.connect_after(self._main_channel, "channel-event", self._main_channel_event_cb)
            _SIGS.connect_after(
                self._main_channel, "notify::agent-connected", self._agent_connected_cb
            )

        elif isinstance(channel, SpiceClientGLib.DisplayChannel) and not self._display:
            channel_id = channel.get_property("channel-id")

            if channel_id != 0:  # pragma: no cover
                log.debug("Spice multi-head unsupported")
                return

            display = SpiceClientGtk.Display.new(self._spice_session, channel_id)
            self._set_display(display)

            _SIGS.connect_after(channel, "display-primary-create", self._display_primary_create_cb)
            self.emit("connected")

        elif (
            type(channel) in [SpiceClientGLib.PlaybackChannel, SpiceClientGLib.RecordChannel]
            and not self._audio
        ):
            # It's unclear why we need this audio handle, but it
            # does matter:
            # https://bugzilla.redhat.com/show_bug.cgi?id=1881080
            self._audio = SpiceClientGLib.Audio.get(self._spice_session, None)

    def _agent_connected_cb(self, src, val):
        self.emit("agent-connected")  # pragma: no cover

    def _display_primary_create_cb(self, *args):
        self._set_desktop_resolution(args[2], args[3])

    ################################
    # Internal API implementations #
    ################################

    def _close(self):
        if self._spice_session is not None:
            _SIGS.disconnect_obj_signals(self._spice_session)
            self._spice_session.disconnect()
        self._spice_session = None
        self._audio = None

        for channel in self._channels:
            _SIGS.disconnect_obj_signals(channel)
        self._channels = None

        self._main_channel = None

        _SIGS.disconnect_obj_signals(self._usbdev_manager)
        self._usbdev_manager = None

    def _is_open(self):
        return self._spice_session is not None

    def _set_grab_keys(self, keys):
        seq = SpiceClientGtk.GrabSequence.new(keys)
        self._display.set_grab_keys(seq)

    def _send_keys(self, keys):
        return self._display.send_keys(
            [Gdk.keyval_from_name(k) for k in keys], SpiceClientGtk.DisplayKeyEvent.CLICK
        )

    def _has_agent(self):
        if not self._main_channel:
            return False  # pragma: no cover
        return self._main_channel.get_property("agent-connected")

    def _open_host(self):
        host, port, tlsport = self._ginfo.get_conn_host()
        self._create_spice_session()

        log.debug("Spice connecting to host=%s port=%s tlsport=%s", host, port, tlsport)
        self._spice_session.set_property("host", str(host))
        if port:
            self._spice_session.set_property("port", str(port))
        if tlsport:
            self._spice_session.set_property("tls-port", str(tlsport))

        self._spice_session.connect()

    def _open_fd(self, fd):
        self._create_spice_session()
        self._spice_session.open_fd(fd)

    def _set_username(self, cred):
        ignore = cred  # pragma: no cover

    def _set_password(self, cred):
        self._spice_session.set_property("password", cred)
        fd = self._get_fd_for_open()
        if fd is not None:
            self._spice_session.open_fd(fd)
        else:
            self._spice_session.connect()  # pragma: no cover

    def _get_scaling(self):
        if self._display:
            return self._display.get_property("scaling")

    def _set_scaling(self, scaling):
        if self._display:
            self._display.set_property("scaling", scaling)

    def _set_resizeguest(self, val):
        if self._display:
            self._display.set_property("resize-guest", val)

    def _get_resizeguest(self):
        if self._display:
            return self._display.get_property("resize-guest")
        return False  # pragma: no cover

    def _get_resizeguest_warning(self):
        if not self._has_agent():
            return _("Guest agent is not available.")

    def _usbdev_redirect_error(
        self, spice_usbdev_widget, spice_usb_device, errstr
    ):  # pragma: no cover
        ignore = spice_usbdev_widget
        ignore = spice_usb_device
        self.emit("usb-redirect-error", errstr)

    def _get_usb_widget(self):
        if not self._spice_session:
            return  # pragma: no cover

        usbwidget = SpiceClientGtk.UsbDeviceWidget.new(self._spice_session, None)
        usbwidget.connect("connect-failed", self._usbdev_redirect_error)
        return usbwidget

    def _has_usb_redirection(self):
        if not self._spice_session or not self._usbdev_manager:
            return False  # pragma: no cover
        return True

    def _get_preferred_size(self):
        if not self._display or self._get_scaling() or self._get_resizeguest():
            return self._desktop_resolution

        # When scaling and resizeguest disabled, this usually matches the
        # VM resolution. But when host desktop scaling is used, spice will
        # intentionally have different values here.
        w = self._display.get_preferred_width()[1]
        h = self._display.get_preferred_height()[1]
        return (w, h)



# Copyright (C) 2018 VMware, Inc.
#
# This work is licensed under the GNU GPLv2 or later.
# See the COPYING file in the top-level directory.

from .device import Device
from ..xmlbuilder import XMLProperty


class DeviceVsock(Device):
    XML_NAME = "vsock"

    model = XMLProperty("./@model")
    auto_cid = XMLProperty("./cid/@auto", is_yesno=True)
    cid = XMLProperty("./cid/@address", is_int=True)

    ##################
    # Default config #
    ##################

    def set_defaults(self, guest):
        if not self.model:
            self.model = "virtio"

        if self.auto_cid is None and self.cid is None:
            self.auto_cid = True
